@@ -2,8 +2,6 @@
 
 # v2: replace with src/awkward/_v2/_util.py
 
-from __future__ import absolute_import
-
 import re
 import sys
 import os
@@ -12,30 +10,18 @@ import warnings
 import itertools
 import numbers
 
-try:
-    from collections.abc import Mapping
-    from collections.abc import MutableMapping
-except ImportError:
-    from collections import Mapping
-    from collections import MutableMapping
+from collections.abc import Mapping
+from collections.abc import MutableMapping
+from collections.abc import Iterable
+from collections.abc import Sized
 
 import awkward as ak
 
 np = ak.nplike.NumpyMetadata.instance()
 
-py27 = sys.version_info[0] < 3
-py35 = sys.version_info[0] == 3 and sys.version_info[1] <= 5
-py36 = sys.version_info[0] == 3 and sys.version_info[1] <= 6
-
 win = os.name == "nt"
 bits32 = ak.nplike.numpy.iinfo(np.intp).bits == 32
 
-# to silence flake8 F821 errors
-if py27:
-    unicode = eval("unicode")
-else:
-    unicode = None
-
 # matches include/awkward/common.h
 kMaxInt8 = 127  # 2**7  - 1
 kMaxUInt8 = 255  # 2**8  - 1
@@ -53,14 +39,18 @@ def is_file_path(x):
         return False
 
 
+def is_sized_iterable(obj):
+    return isinstance(obj, Iterable) and isinstance(obj, Sized)
+
+
 def isint(x):
     """
     Returns True if and only if ``x`` is an integer (including NumPy, not
     including bool).
     """
-    return isinstance(
-        x, (int, numbers.Integral, np.integer, ak._v2._typetracer.Interval)
-    ) and not isinstance(x, (bool, np.bool_))
+    return isinstance(x, (int, numbers.Integral, np.integer)) and not isinstance(
+        x, (bool, np.bool_)
+    )
 
 
 def isnum(x):
@@ -75,12 +65,10 @@ def isnum(x):
 
 def isstr(x):
     """
-    Returns True if and only if ``x`` is a string (including Python 2 unicode).
+    Returns True if and only if ``x`` is a string. Mostly needed for Python 2
+    compatibility in the past, but still mirrors isnum.
     """
-    if py27:
-        return isinstance(x, (bytes, unicode))
-    else:
-        return isinstance(x, str)
+    return isinstance(x, str)
 
 
 def exception_suffix(filename):
@@ -114,7 +102,7 @@ def deprecate(
         date = ""
     else:
         date = " (target date: " + date + ")"
-    warning = """In version {0}{1}, this will be {2}.
+    warning = """In version {}{}, this will be {}.
 
 To raise these warnings as errors (and get stack traces to find out where they're called), run
 
@@ -123,14 +111,14 @@ To raise these warnings as errors (and get stack traces to find out where they'r
 
 after the first `import awkward` or use `@pytest.mark.filterwarnings("error:::awkward.*")` in pytest.
 
-Issue: {3}.""".format(
+Issue: {}.""".format(
         version, date, will_be, message
     )
     warnings.warn(warning, category, stacklevel=stacklevel + 1)
 
 
 # Sentinel object for catching pass-through values
-class MISSING(object):
+class MISSING:
     pass
 
 
@@ -225,8 +213,7 @@ class Behavior(Mapping):
                 yield n, x
 
     def __iter__(self):
-        for x in self.keys():
-            yield x
+        yield from self.keys()
 
     def __len__(self):
         return len(set(self.defaults) | set(self.overrides))
@@ -236,17 +223,17 @@ def arrayclass(layout, behavior):
     layout = ak.partition.first(layout)
     behavior = Behavior(ak.behavior, behavior)
     arr = layout.parameter("__array__")
-    if isinstance(arr, str) or (py27 and isinstance(arr, unicode)):
+    if isinstance(arr, str):
         cls = behavior[arr]
         if isinstance(cls, type) and issubclass(cls, ak.highlevel.Array):
             return cls
     rec = layout.parameter("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         cls = behavior[".", rec]
         if isinstance(cls, type) and issubclass(cls, ak.highlevel.Array):
             return cls
     deeprec = layout.purelist_parameter("__record__")
-    if isinstance(deeprec, str) or (py27 and isinstance(deeprec, unicode)):
+    if isinstance(deeprec, str):
         cls = behavior["*", deeprec]
         if isinstance(cls, type) and issubclass(cls, ak.highlevel.Array):
             return cls
@@ -270,11 +257,11 @@ def custom_broadcast(layout, behavior):
     layout = ak.partition.first(layout)
     behavior = Behavior(ak.behavior, behavior)
     custom = layout.parameter("__array__")
-    if not (isinstance(custom, str) or (py27 and isinstance(custom, unicode))):
+    if not isinstance(custom, str):
         custom = layout.parameter("__record__")
-    if not (isinstance(custom, str) or (py27 and isinstance(custom, unicode))):
+    if not isinstance(custom, str):
         custom = layout.purelist_parameter("__record__")
-    if isinstance(custom, str) or (py27 and isinstance(custom, unicode)):
+    if isinstance(custom, str):
         for key, fcn in behavior.items():
             if (
                 isinstance(key, tuple)
@@ -289,17 +276,17 @@ def custom_broadcast(layout, behavior):
 def numba_array_typer(layouttype, behavior):
     behavior = Behavior(ak.behavior, behavior)
     arr = layouttype.parameters.get("__array__")
-    if isinstance(arr, str) or (py27 and isinstance(arr, unicode)):
+    if isinstance(arr, str):
         typer = behavior["__numba_typer__", arr]
         if callable(typer):
             return typer
     rec = layouttype.parameters.get("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         typer = behavior["__numba_typer__", ".", rec]
         if callable(typer):
             return typer
     deeprec = layouttype.parameters.get("__record__")
-    if isinstance(deeprec, str) or (py27 and isinstance(deeprec, unicode)):
+    if isinstance(deeprec, str):
         typer = behavior["__numba_typer__", "*", deeprec]
         if callable(typer):
             return typer
@@ -309,17 +296,17 @@ def numba_array_typer(layouttype, behavior):
 def numba_array_lower(layouttype, behavior):
     behavior = Behavior(ak.behavior, behavior)
     arr = layouttype.parameters.get("__array__")
-    if isinstance(arr, str) or (py27 and isinstance(arr, unicode)):
+    if isinstance(arr, str):
         lower = behavior["__numba_lower__", arr]
         if callable(lower):
             return lower
     rec = layouttype.parameters.get("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         lower = behavior["__numba_lower__", ".", rec]
         if callable(lower):
             return lower
     deeprec = layouttype.parameters.get("__record__")
-    if isinstance(deeprec, str) or (py27 and isinstance(deeprec, unicode)):
+    if isinstance(deeprec, str):
         lower = behavior["__numba_lower__", "*", deeprec]
         if callable(lower):
             return lower
@@ -330,7 +317,7 @@ def recordclass(layout, behavior):
     layout = ak.partition.first(layout)
     behavior = Behavior(ak.behavior, behavior)
     rec = layout.parameter("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         cls = behavior[rec]
         if isinstance(cls, type) and issubclass(cls, ak.highlevel.Record):
             return cls
@@ -345,8 +332,8 @@ def typestrs(behavior):
             isinstance(key, tuple)
             and len(key) == 2
             and key[0] == "__typestr__"
-            and (isinstance(key[1], str) or (py27 and isinstance(key[1], unicode)))
-            and (isinstance(typestr, str) or (py27 and isinstance(typestr, unicode)))
+            and isinstance(key[1], str)
+            and isinstance(typestr, str)
         ):
             out[key[1]] = typestr
     return out
@@ -370,7 +357,7 @@ def gettypestr(parameters, typestrs):
 def numba_record_typer(layouttype, behavior):
     behavior = Behavior(ak.behavior, behavior)
     rec = layouttype.parameters.get("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         typer = behavior["__numba_typer__", rec]
         if callable(typer):
             return typer
@@ -380,7 +367,7 @@ def numba_record_typer(layouttype, behavior):
 def numba_record_lower(layouttype, behavior):
     behavior = Behavior(ak.behavior, behavior)
     rec = layouttype.parameters.get("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         lower = behavior["__numba_lower__", rec]
         if callable(lower):
             return lower
@@ -409,7 +396,7 @@ def overload(behavior, signature):
 def numba_attrs(layouttype, behavior):
     behavior = Behavior(ak.behavior, behavior)
     rec = layouttype.parameters.get("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         for key, typer in behavior.items():
             if (
                 isinstance(key, tuple)
@@ -424,7 +411,7 @@ def numba_attrs(layouttype, behavior):
 def numba_methods(layouttype, behavior):
     behavior = Behavior(ak.behavior, behavior)
     rec = layouttype.parameters.get("__record__")
-    if isinstance(rec, str) or (py27 and isinstance(rec, unicode)):
+    if isinstance(rec, str):
         for key, typer in behavior.items():
             if (
                 isinstance(key, tuple)
@@ -443,7 +430,7 @@ def numba_unaryops(unaryop, left, behavior):
 
     if isinstance(left, ak._connect._numba.layout.ContentType):
         left = left.parameters.get("__record__")
-        if not (isinstance(left, str) or (py27 and isinstance(left, unicode))):
+        if not isinstance(left, str):
             done = True
 
     if not done:
@@ -465,12 +452,12 @@ def numba_binops(binop, left, right, behavior):
 
     if isinstance(left, ak._connect._numba.layout.ContentType):
         left = left.parameters.get("__record__")
-        if not (isinstance(left, str) or (py27 and isinstance(left, unicode))):
+        if not isinstance(left, str):
             done = True
 
     if isinstance(right, ak._connect._numba.layout.ContentType):
         right = right.parameters.get("__record__")
-        if not isinstance(right, str) and not (py27 and isinstance(right, unicode)):
+        if not isinstance(right, str):
             done = True
 
     if not done:
@@ -539,8 +526,7 @@ def maybe_wrap_like(content, array, behavior, highlevel):
 
 def extra(args, kwargs, defaults):
     out = []
-    for i in range(len(defaults)):
-        name, default = defaults[i]
+    for i, (name, default) in enumerate(defaults):
         if i < len(args):
             out.append(args[i])
         elif name in kwargs:
@@ -566,7 +552,7 @@ def key2index(keys, key):
 
     if attempt is None:
         raise ValueError(
-            "key {0} not found in record".format(repr(key)) + exception_suffix(__file__)
+            f"key {key!r} not found in record" + exception_suffix(__file__)
         )
     else:
         return attempt
@@ -626,8 +612,7 @@ def completely_flatten(array):
 
     else:
         raise RuntimeError(
-            "cannot completely flatten: {0}".format(type(array))
-            + exception_suffix(__file__)
+            f"cannot completely flatten: {type(array)}" + exception_suffix(__file__)
         )
 
 
@@ -649,8 +634,8 @@ def broadcast_and_apply(  # noqa: C901
         for x in inputs[1:]:
             if len(x) != length:
                 raise ValueError(
-                    "cannot broadcast {0} of length {1} with {2} of "
-                    "length {3}".format(
+                    "cannot broadcast {} of length {} with {} of "
+                    "length {}".format(
                         type(inputs[0]).__name__, length, type(x).__name__, len(x)
                     )
                     + exception_suffix(__file__)
@@ -705,8 +690,7 @@ def broadcast_and_apply(  # noqa: C901
                     return False
             elif isinstance(x, ak.layout.Content):
                 return False
-        else:
-            return True
+        return True
 
     def apply(inputs, depth, user):
         nplike = ak.nplike.of(*inputs)
@@ -809,8 +793,8 @@ def broadcast_and_apply(  # noqa: C901
                         length = len(tagslist[-1])
                     elif length != len(tagslist[-1]):
                         raise ValueError(
-                            "cannot broadcast UnionArray of length {0} "
-                            "with UnionArray of length {1}".format(
+                            "cannot broadcast UnionArray of length {} "
+                            "with UnionArray of length {}".format(
                                 length, len(tagslist[-1])
                             )
                             + exception_suffix(__file__)
@@ -906,7 +890,7 @@ def broadcast_and_apply(  # noqa: C901
                 for x in inputs
             ):
                 maxsize = max(
-                    [x.size for x in inputs if isinstance(x, ak.layout.RegularArray)]
+                    x.size for x in inputs if isinstance(x, ak.layout.RegularArray)
                 )
                 for x in inputs:
                     if isinstance(x, ak.layout.RegularArray):
@@ -930,7 +914,7 @@ def broadcast_and_apply(  # noqa: C901
                         else:
                             raise ValueError(
                                 "cannot broadcast RegularArray of size "
-                                "{0} with RegularArray of size {1}".format(
+                                "{} with RegularArray of size {}".format(
                                     x.size, maxsize
                                 )
                                 + exception_suffix(__file__)
@@ -938,14 +922,20 @@ def broadcast_and_apply(  # noqa: C901
                     else:
                         nextinputs.append(x)
 
-                maxlen = max(
-                    [len(x) for x in nextinputs if isinstance(x, ak.layout.Content)]
-                )
                 outcontent = apply(nextinputs, depth + 1, user)
                 assert isinstance(outcontent, tuple)
 
+                length = None
+                for x in inputs:
+                    if isinstance(x, ak.layout.Content):
+                        if length is None:
+                            length = len(x)
+                        else:
+                            assert length == len(x)
+                assert length is not None
+
                 return tuple(
-                    ak.layout.RegularArray(x, maxsize, maxlen) for x in outcontent
+                    ak.layout.RegularArray(x, maxsize, length) for x in outcontent
                 )
 
             elif not all_same_offsets(nplike, inputs):
@@ -1063,7 +1053,7 @@ def broadcast_and_apply(  # noqa: C901
                     )
                 else:
                     raise AssertionError(
-                        "unexpected offsets, starts: {0} {1}".format(
+                        "unexpected offsets, starts: {} {}".format(
                             type(offsets), type(starts)
                         )
                         + exception_suffix(__file__)
@@ -1086,7 +1076,7 @@ def broadcast_and_apply(  # noqa: C901
                     elif set(keys) != set(x.keys()):
                         raise ValueError(
                             "cannot broadcast records because keys don't "
-                            "match:\n    {0}\n    {1}".format(
+                            "match:\n    {}\n    {}".format(
                                 ", ".join(sorted(keys)), ", ".join(sorted(x.keys()))
                             )
                             + exception_suffix(__file__)
@@ -1095,8 +1085,8 @@ def broadcast_and_apply(  # noqa: C901
                         length = len(x)
                     elif length != len(x):
                         raise ValueError(
-                            "cannot broadcast RecordArray of length {0} "
-                            "with RecordArray of length {1}".format(length, len(x))
+                            "cannot broadcast RecordArray of length {} "
+                            "with RecordArray of length {}".format(length, len(x))
                             + exception_suffix(__file__)
                         )
                     if not x.istuple:
@@ -1128,7 +1118,7 @@ def broadcast_and_apply(  # noqa: C901
 
         else:
             raise ValueError(
-                "cannot broadcast: {0}".format(", ".join(repr(type(x)) for x in inputs))
+                "cannot broadcast: {}".format(", ".join(repr(type(x)) for x in inputs))
                 + exception_suffix(__file__)
             )
 
@@ -1315,8 +1305,7 @@ def recursive_walk(layout, apply, args=(), depth=1, materialize=False):
 
     else:
         raise AssertionError(
-            "unrecognized Content type: {0}".format(type(layout))
-            + exception_suffix(__file__)
+            f"unrecognized Content type: {type(layout)}" + exception_suffix(__file__)
         )
 
 
@@ -1515,8 +1504,7 @@ def transform_child_layouts(transform, layout, depth, user=None, keep_parameters
 
     else:
         raise AssertionError(
-            "unrecognized Content type: {0}".format(type(layout))
-            + exception_suffix(__file__)
+            f"unrecognized Content type: {type(layout)}" + exception_suffix(__file__)
         )
 
 
@@ -1619,7 +1607,7 @@ def minimally_touching_string(limit_length, layout, behavior):
             elif isinstance(x, (np.datetime64, np.timedelta64)):
                 yield space + str(x)
             elif isinstance(x, (float, np.floating)):
-                yield space + "{0:.3g}".format(x)
+                yield space + f"{x:.3g}"
             else:
                 yield space + repr(x)
 
@@ -1686,13 +1674,12 @@ def minimally_touching_string(limit_length, layout, behavior):
             elif isinstance(x, (np.datetime64, np.timedelta64)):
                 yield str(x) + space
             elif isinstance(x, (float, np.floating)):
-                yield "{0:.3g}".format(x) + space
+                yield f"{x:.3g}" + space
             else:
                 yield repr(x) + space
 
     def forever(iterable):
-        for token in iterable:
-            yield token
+        yield from iterable
         while True:
             yield None
 
