@@ -1,7 +1,13 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
+from __future__ import annotations
+
 import copy
+import enum
+import functools
 import itertools
+from typing import Any, Callable, Dict, List, Union
+from collections.abc import Sequence
 
 import awkward as ak
 from awkward._v2.contents.content import Content  # noqa: F401
@@ -149,6 +155,194 @@ def all_same_offsets(nplike, inputs):
     return True
 
 
+# TODO: move to _util or another module
+class Sentinel:
+    """A class for implementing sentinel types"""
+
+    def __init__(self, name, module=None):
+        self._name = name
+        self._module = module
+
+    def __repr__(self):
+        if self._module is not None:
+            return f"{self._module}.{self._name}"
+        else:
+            return f"{self._name}"
+
+
+NO_PARAMETERS = Sentinel("NO_PARAMETERS", __name__)
+
+
+class BroadcastParameterRule(str, enum.Enum):
+    """Behaviour for parameter coalescence during broadcasting."""
+
+    INTERSECT = "intersect"
+    ONE_TO_ONE = "one_to_one"
+    ALL_OR_NOTHING = "all_or_nothing"
+    NONE = "none"
+
+
+BroadcastParameterFactory = Callable[[int], List[Union[Dict[str, Any], None]]]
+
+
+def _parameters_of(obj: Any, default: Any = NO_PARAMETERS) -> Any:
+    """
+    Args:
+        obj: #ak._v2.contents.Content that holds parameters, or object
+        default: value to return if obj is not an #ak._v2.contents.Content
+
+    Return the parameters of an object if it is a #ak._v2.contents.Content;
+    otherwise, return a default value.
+    """
+    if isinstance(obj, ak._v2.contents.Content):
+        return obj._parameters
+    else:
+        return default
+
+
+def all_or_nothing_parameters_factory(
+    inputs: Sequence,
+) -> BroadcastParameterFactory:
+    """
+    Args:
+        inputs: sequence of #ak._v2.contents.Content or other objects
+
+    Return a callable that creates an appropriately sized list of parameter objects.
+    The parameter objects within this list are built using an "all or nothing rule":
+
+    If the parameters of all the given contents are equal, then the first content's
+    parameters are repeated, i.e. `[parameters, parameters, ...]`. Otherwise, a list
+    of Nones is returned, i.e. `[None, None, ...]`.
+    """
+    input_parameters = [
+        p for p in (_parameters_of(c) for c in inputs) if p is not NO_PARAMETERS
+    ]
+
+    parameters = None
+    if len(input_parameters) > 0:
+        # All parameters must match this first layout's parameters
+        first_parameters = input_parameters[0]
+        # Ensure all parameters match, or set parameters to None
+        for other_parameters in input_parameters[1:]:
+            if not ak._v2.forms.form._parameters_equal(
+                first_parameters, other_parameters
+            ):
+                break
+        else:
+            parameters = first_parameters
+
+    def apply(n_outputs: int) -> list[dict[str, Any] | None]:
+        # NB: we don't make unique copies here, so let's hope everyone
+        # is well-behaved downstream!
+        return [parameters] * n_outputs
+
+    return apply
+
+
+def intersection_parameters_factory(
+    inputs: Sequence,
+) -> BroadcastParameterFactory:
+    """
+    Args:
+        inputs: sequence of #ak._v2.contents.Content or other objects
+
+    Return a callable that creates an appropriately sized list of parameter objects.
+    The parameter objects within this list are built using an "intersection rule":
+
+    The intersection of `content._parameters.items()` for each content is computed.
+    If any parameter dictionaries are None, then a list of Nones is returned, i.e.
+    `[None, None, ...]`; otherwise, the computed parameter dictionary is repeated,
+    i.e. `[parameters, parameters, ...]`.
+    """
+    input_parameters = [
+        p for p in (_parameters_of(c) for c in inputs) if p is not NO_PARAMETERS
+    ]
+
+    intersected_parameters = None
+    parameters_to_intersect = []
+    # Build a list of set-like dict.items() views.
+    # If we encounter None-parameters, then we stop early
+    # as there can be no intersection.
+    for parameters in input_parameters:
+        if ak._v2.forms.form._parameters_is_empty(parameters):
+            break
+        else:
+            parameters_to_intersect.append(parameters)
+    # Otherwise, build the intersected parameter dict
+    else:
+        intersected_parameters = functools.reduce(
+            ak._v2.forms.form._parameters_intersect, parameters_to_intersect
+        )
+
+    def apply(n_outputs: int) -> list[dict[str, Any] | None]:
+        # NB: we don't make unique copies here, so let's hope everyone
+        # is well-behaved downstream!
+        return [intersected_parameters] * n_outputs
+
+    return apply
+
+
+def one_to_one_parameters_factory(
+    inputs: Sequence,
+) -> BroadcastParameterFactory:
+    """
+    Args:
+        inputs: sequence of #ak._v2.contents.Content or other objects
+
+    Return a callable that creates an appropriately sized list of parameter objects.
+    The parameter objects within this list are built using a "one-to-one rule":
+
+    The requested number of outputs is compared against the length of the given
+    `inputs`. If the two values match, then a list of parameter objects is returned,
+    where each element of the returned list corresponds to the parameters of the
+    content at the same position in the `inputs` sequence. If the length of the
+    given contents does not match the requested list length, a ValueError is raised.
+    """
+    # Find the parameters of the inputs, with None values for non-Contents
+    input_parameters = [_parameters_of(c, default=None) for c in inputs]
+
+    def apply(n_outputs) -> list[dict[str, Any] | None]:
+        if n_outputs != len(inputs):
+            raise ak._v2._util.error(
+                ValueError(
+                    "cannot follow one-to-one parameter broadcasting rule for actions "
+                    "which change the number of outputs."
+                )
+            )
+        return input_parameters
+
+    return apply
+
+
+def none_parameters_factory(
+    inputs: Sequence,
+) -> BroadcastParameterFactory:
+    """
+    Args:
+        inputs: sequence of #ak._v2.contents.Content or other objects
+
+    Return a callable that creates an appropriately sized list of parameter objects.
+    The parameter objects within this list are built using an "all or nothing rule":
+
+    A list of Nones is returned, with a length corresponding to the requested number of
+    outputs, i.e. `[None, None, ...]`.
+    """
+
+    def apply(n_outputs: int) -> list[dict[str, Any] | None]:
+        return [None] * n_outputs
+
+    return apply
+
+
+# Mapping from rule enum values to factory implementations
+BROADCAST_RULE_TO_FACTORY_IMPL = {
+    BroadcastParameterRule.INTERSECT: intersection_parameters_factory,
+    BroadcastParameterRule.ALL_OR_NOTHING: all_or_nothing_parameters_factory,
+    BroadcastParameterRule.ONE_TO_ONE: one_to_one_parameters_factory,
+    BroadcastParameterRule.NONE: none_parameters_factory,
+}
+
+
 def apply_step(
     nplike, inputs, action, depth, depth_context, lateral_context, behavior, options
 ):
@@ -198,6 +392,19 @@ def apply_step(
     if nplike.known_shape:
         checklength([x for x in inputs if isinstance(x, Content)], options)
 
+    # Load the parameter broadcasting rule implementation
+    rule = options["broadcast_parameters_rule"]
+    try:
+        parameters_factory_impl = BROADCAST_RULE_TO_FACTORY_IMPL[rule]
+    except KeyError:
+        raise ak._v2._util.error(
+            ValueError(
+                f"`broadcast_parameters_rule` should be one of {[str(x) for x in BroadcastParameterRule]}, "
+                f"but this routine received `{rule}`"
+            )
+        ) from None
+    parameters_factory = parameters_factory_impl(inputs)
+
     # This whole function is one big switch statement.
     def continuation():
         # Any EmptyArrays?
@@ -366,11 +573,15 @@ def apply_step(
 
                 assert numoutputs is not None
 
+            parameters = parameters_factory(numoutputs)
             return tuple(
                 UnionArray(
-                    Index8(tags), Index64(index), [x[i] for x in outcontents]
+                    Index8(tags),
+                    Index64(index),
+                    [x[i] for x in outcontents],
+                    parameters=p,
                 ).simplify_uniontype()
-                for i in range(numoutputs)
+                for i, p in enumerate(parameters)
             )
 
         # Any option-types?
@@ -434,8 +645,10 @@ def apply_step(
                 options,
             )
             assert isinstance(outcontent, tuple)
+            parameters = parameters_factory(len(outcontent))
             return tuple(
-                IndexedOptionArray(index, x).simplify_optiontype() for x in outcontent
+                IndexedOptionArray(index, x, parameters=p).simplify_optiontype()
+                for x, p in zip(outcontent, parameters)
             )
 
         # Any list-types?
@@ -512,7 +725,11 @@ def apply_step(
                     options,
                 )
                 assert isinstance(outcontent, tuple)
-                return tuple(RegularArray(x, maxsize, length) for x in outcontent)
+                parameters = parameters_factory(len(outcontent))
+                return tuple(
+                    RegularArray(x, maxsize, length, parameters=p)
+                    for x, p in zip(outcontent, parameters)
+                )
 
             elif not nplike.known_data or not nplike.known_shape:
                 offsets = None
@@ -551,8 +768,11 @@ def apply_step(
                     options,
                 )
                 assert isinstance(outcontent, tuple)
-
-                return tuple(ListOffsetArray(offsets, x) for x in outcontent)
+                parameters = parameters_factory(len(outcontent))
+                return tuple(
+                    ListOffsetArray(offsets, x, parameters=p)
+                    for x, p in zip(outcontent, parameters)
+                )
 
             # Not all regular, but all same offsets?
             # Optimization: https://github.com/scikit-hep/awkward-1.0/issues/442
@@ -588,16 +808,21 @@ def apply_step(
                     options,
                 )
                 assert isinstance(outcontent, tuple)
+                parameters = parameters_factory(len(outcontent))
 
                 if isinstance(offsets, Index):
                     return tuple(
-                        ListOffsetArray(offsets, x).toListOffsetArray64(False)
-                        for x in outcontent
+                        ListOffsetArray(offsets, x, parameters=p).toListOffsetArray64(
+                            False
+                        )
+                        for x, p in zip(outcontent, parameters)
                     )
                 elif isinstance(starts, Index) and isinstance(stops, Index):
                     return tuple(
-                        ListArray(starts, stops, x).toListOffsetArray64(False)
-                        for x in outcontent
+                        ListArray(starts, stops, x, parameters=p).toListOffsetArray64(
+                            False
+                        )
+                        for x, p in zip(outcontent, parameters)
                     )
                 else:
                     raise ak._v2._util.error(
@@ -617,7 +842,11 @@ def apply_step(
                     for x in inputs
                 ]
 
-                first, secondround = None, False
+                # Find first list without custom broadcasting which will define the broadcast offsets
+                # Don't consider RegularArray (handle case of 1-sized regular broadcasting)
+                # Do this to prioritise non-custom broadcasting
+                first = None
+
                 for x, fcn in zip(inputs, fcns):
                     if (
                         isinstance(x, listtypes)
@@ -627,6 +856,10 @@ def apply_step(
                         first = x
                         break
 
+                # Did we fail to find a list without custom broadcasting?
+                secondround = False
+                # If we failed to find an irregular, non-custom broadcasting list,
+                # continue search for *any* list.
                 if first is None:
                     secondround = True
                     for x in inputs:
@@ -664,8 +897,12 @@ def apply_step(
                     options,
                 )
                 assert isinstance(outcontent, tuple)
+                parameters = parameters_factory(len(outcontent))
 
-                return tuple(ListOffsetArray(offsets, x) for x in outcontent)
+                return tuple(
+                    ListOffsetArray(offsets, x, parameters=p)
+                    for x, p in zip(outcontent, parameters)
+                )
 
         # Any RecordArrays?
         elif any(isinstance(x, RecordArray) for x in inputs):
@@ -724,9 +961,12 @@ def apply_step(
 
             return tuple(
                 RecordArray(
-                    [x[i] for x in outcontents], None if istuple else fields, length
+                    [x[i] for x in outcontents],
+                    None if istuple else fields,
+                    length,
+                    parameters=p,
                 )
-                for i in range(numoutputs)
+                for i, p in enumerate(parameters_factory(numoutputs))
             )
 
         else:
@@ -769,6 +1009,7 @@ def broadcast_and_apply(
     numpy_to_regular=False,
     regular_to_jagged=False,
     function_name=None,
+    broadcast_parameters_rule=BroadcastParameterRule.INTERSECT,
 ):
     nplike = ak.nplike.of(*inputs)
     isscalar = []
@@ -787,6 +1028,7 @@ def broadcast_and_apply(
             "numpy_to_regular": numpy_to_regular,
             "regular_to_jagged": regular_to_jagged,
             "function_name": function_name,
+            "broadcast_parameters_rule": broadcast_parameters_rule,
         },
     )
     assert isinstance(out, tuple)
