@@ -7,10 +7,8 @@ from collections.abc import Iterable
 
 import awkward as ak
 from awkward._v2.record import Record
-from awkward._v2._slicing import NestedIndexError
-from awkward._v2.contents.content import Content
+from awkward._v2.contents.content import Content, unset
 from awkward._v2.forms.recordform import RecordForm
-from awkward._v2.forms.form import _parameters_equal
 
 np = ak.nplike.NumpyMetadata.instance()
 numpy = ak.nplike.Numpy.instance()
@@ -19,6 +17,35 @@ numpy = ak.nplike.Numpy.instance()
 class RecordArray(Content):
     is_RecordType = True
 
+    def copy(
+        self,
+        contents=unset,
+        fields=unset,
+        length=unset,
+        identifier=unset,
+        parameters=unset,
+        nplike=unset,
+    ):
+        return RecordArray(
+            self._contents if contents is unset else contents,
+            self._fields if fields is unset else fields,
+            self._length if length is unset else length,
+            self._identifier if identifier is unset else identifier,
+            self._parameters if parameters is unset else parameters,
+            self._nplike if nplike is unset else nplike,
+        )
+
+    def __copy__(self):
+        return self.copy()
+
+    def __deepcopy__(self, memo):
+        return self.copy(
+            contents=[copy.deepcopy(x, memo) for x in self._contents],
+            fields=copy.deepcopy(self._fields, memo),
+            identifier=copy.deepcopy(self._identifier, memo),
+            parameters=copy.deepcopy(self._parameters, memo),
+        )
+
     def __init__(
         self,
         contents,
@@ -29,18 +56,22 @@ class RecordArray(Content):
         nplike=None,
     ):
         if not isinstance(contents, Iterable):
-            raise TypeError(
-                "{} 'contents' must be iterable, not {}".format(
-                    type(self).__name__, repr(contents)
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} 'contents' must be iterable, not {}".format(
+                        type(self).__name__, repr(contents)
+                    )
                 )
             )
         if not isinstance(contents, list):
             contents = list(contents)
 
         if len(contents) == 0 and length is None:
-            raise TypeError(
-                "{} if len(contents) == 0, a 'length' must be specified".format(
-                    type(self).__name__
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} if len(contents) == 0, a 'length' must be specified".format(
+                        type(self).__name__
+                    )
                 )
             )
         elif length is None:
@@ -54,22 +85,28 @@ class RecordArray(Content):
         if not isinstance(length, ak._v2._typetracer.UnknownLengthType) and not (
             ak._util.isint(length) and length >= 0
         ):
-            raise TypeError(
-                "{} 'length' must be a non-negative integer or None, not {}".format(
-                    type(self).__name__, repr(length)
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} 'length' must be a non-negative integer or None, not {}".format(
+                        type(self).__name__, repr(length)
+                    )
                 )
             )
         for content in contents:
             if not isinstance(content, Content):
-                raise TypeError(
-                    "{} all 'contents' must be Content subclasses, not {}".format(
-                        type(self).__name__, repr(content)
+                raise ak._v2._util.error(
+                    TypeError(
+                        "{} all 'contents' must be Content subclasses, not {}".format(
+                            type(self).__name__, repr(content)
+                        )
                     )
                 )
             if content.length < length:
-                raise ValueError(
-                    "{} len(content) ({}) must be >= length ({}) for all 'contents'".format(
-                        type(self).__name__, content.length, length
+                raise ak._v2._util.error(
+                    ValueError(
+                        "{} len(content) ({}) must be >= length ({}) for all 'contents'".format(
+                            type(self).__name__, content.length, length
+                        )
                     )
                 )
 
@@ -77,21 +114,27 @@ class RecordArray(Content):
             if not isinstance(fields, list):
                 fields = list(fields)
             if not all(ak._util.isstr(x) for x in fields):
-                raise TypeError(
-                    "{} 'fields' must all be strings, not {}".format(
-                        type(self).__name__, repr(fields)
+                raise ak._v2._util.error(
+                    TypeError(
+                        "{} 'fields' must all be strings, not {}".format(
+                            type(self).__name__, repr(fields)
+                        )
                     )
                 )
             if not len(contents) == len(fields):
-                raise ValueError(
-                    "{} len(contents) ({}) must be equal to len(fields) ({})".format(
-                        type(self).__name__, len(contents), len(fields)
+                raise ak._v2._util.error(
+                    ValueError(
+                        "{} len(contents) ({}) must be equal to len(fields) ({})".format(
+                            type(self).__name__, len(contents), len(fields)
+                        )
                     )
                 )
         elif fields is not None:
-            raise TypeError(
-                "{} 'fields' must be iterable or None, not {}".format(
-                    type(self).__name__, repr(fields)
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} 'fields' must be iterable or None, not {}".format(
+                        type(self).__name__, repr(fields)
+                    )
                 )
             )
         if nplike is None:
@@ -100,11 +143,13 @@ class RecordArray(Content):
                     nplike = content.nplike
                     break
                 elif nplike is not content.nplike:
-                    raise TypeError(
-                        "{} 'contents' must use the same array library (nplike): {} vs {}".format(
-                            type(self).__name__,
-                            type(nplike).__name__,
-                            type(content.nplike).__name__,
+                    raise ak._v2._util.error(
+                        TypeError(
+                            "{} 'contents' must use the same array library (nplike): {} vs {}".format(
+                                type(self).__name__,
+                                type(nplike).__name__,
+                                type(content.nplike).__name__,
+                            )
                         )
                     )
         if nplike is None:
@@ -253,7 +298,7 @@ class RecordArray(Content):
             where += self.length
 
         if where < 0 or where >= self.length:
-            raise NestedIndexError(self, where)
+            raise ak._v2._util.indexerror(self, where)
         return Record(self, where)
 
     def _getitem_range(self, where):
@@ -326,7 +371,7 @@ class RecordArray(Content):
             self._nplike,
         )
 
-    def _carry(self, carry, allow_lazy, exception):
+    def _carry(self, carry, allow_lazy):
         assert isinstance(carry, ak._v2.index.Index)
 
         if allow_lazy:
@@ -336,27 +381,24 @@ class RecordArray(Content):
                 where = where.copy()
 
             negative = where < 0
-            if self._nplike.any(negative, prefer=False):
+            if self._nplike.index_nplike.any(negative, prefer=False):
                 where[negative] += self._length
 
-            if self._nplike.any(where >= self._length, prefer=False):
-                if issubclass(exception, NestedIndexError):
-                    raise exception(self, where)
-                else:
-                    raise exception("index out of range")
+            if self._nplike.index_nplike.any(where >= self._length, prefer=False):
+                raise ak._v2._util.indexerror(self, where)
 
-            nextindex = ak._v2.index.Index64(where)
+            nextindex = ak._v2.index.Index64(where, nplike=self.nplike)
             return ak._v2.contents.indexedarray.IndexedArray(
                 nextindex,
                 self,
-                self._carry_identifier(carry, exception),
+                self._carry_identifier(carry),
                 None,
                 self._nplike,
             )
 
         else:
             contents = [
-                self.content(i)._carry(carry, allow_lazy, exception)
+                self.content(i)._carry(carry, allow_lazy)
                 for i in range(len(self._contents))
             ]
 
@@ -371,7 +413,7 @@ class RecordArray(Content):
                 contents,
                 self._fields,
                 length,
-                self._carry_identifier(carry, exception),
+                self._carry_identifier(carry),
                 self._parameters,
                 self._nplike,
             )
@@ -430,13 +472,12 @@ class RecordArray(Content):
     def num(self, axis, depth=0):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            single = ak._v2.index.Index64.empty(1, self._nplike)
-            single[0] = self.length
+            npsingle = self._nplike.index_nplike.full((1,), self.length, np.int64)
+            single = ak._v2.index.Index64(npsingle, nplike=self._nplike)
             singleton = ak._v2.contents.numpyarray.NumpyArray(
                 single, None, None, self._nplike
             )
             contents = [singleton] * len(self._contents)
-
             record = ak._v2.contents.recordarray.RecordArray(
                 contents, self._fields, 1, None, self._parameters, self._nplike
             )
@@ -457,11 +498,13 @@ class RecordArray(Content):
     def _offsets_and_flattened(self, axis, depth):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            raise np.AxisError(self, "axis=0 not allowed for flatten")
+            raise ak._v2._util.error(np.AxisError("axis=0 not allowed for flatten"))
 
         elif posaxis == depth + 1:
-            raise ValueError(
-                "arrays of records cannot be flattened (but their contents can be; try a different 'axis')"
+            raise ak._v2._util.error(
+                ValueError(
+                    "arrays of records cannot be flattened (but their contents can be; try a different 'axis')"
+                )
             )
 
         else:
@@ -470,8 +513,10 @@ class RecordArray(Content):
                 trimmed = content._getitem_range(slice(0, self.length))
                 offsets, flattened = trimmed._offsets_and_flattened(posaxis, depth)
                 if self._nplike.known_shape and offsets.length != 0:
-                    raise AssertionError(
-                        "RecordArray content with axis > depth + 1 returned a non-empty offsets from offsets_and_flattened"
+                    raise ak._v2._util.error(
+                        AssertionError(
+                            "RecordArray content with axis > depth + 1 returned a non-empty offsets from offsets_and_flattened"
+                        )
                     )
                 contents.append(flattened)
             offsets = ak._v2.index.Index64.zeros(1, self._nplike, dtype=np.int64)
@@ -487,20 +532,8 @@ class RecordArray(Content):
                 ),
             )
 
-    def mergeable(self, other, mergebool=True):
-        if not _parameters_equal(self._parameters, other._parameters):
-            return False
-
+    def _mergeable(self, other, mergebool):
         if isinstance(
-            other,
-            (
-                ak._v2.contents.emptyarray.EmptyArray,
-                ak._v2.contents.unionarray.UnionArray,
-            ),
-        ):
-            return True
-
-        elif isinstance(
             other,
             (
                 ak._v2.contents.indexedarray.IndexedArray,
@@ -550,12 +583,12 @@ class RecordArray(Content):
         for_each_field = []
         for field in self.contents:
             trimmed = field[0 : self.length]
-            for_each_field.append([field])
+            for_each_field.append([trimmed])
 
         if self.is_tuple:
             for array in headless:
                 parameters = ak._v2._util.merge_parameters(
-                    self._parameters, array._parameters, True
+                    parameters, array._parameters, True
                 )
 
                 if isinstance(array, ak._v2.contents.recordarray.RecordArray):
@@ -565,19 +598,25 @@ class RecordArray(Content):
                                 field = array[self.index_to_field(i)]
                                 for_each_field[i].append(field[0 : array.length])
                         else:
-                            raise ValueError(
-                                "cannot merge tuples with different numbers of fields"
+                            raise ak._v2._util.error(
+                                ValueError(
+                                    "cannot merge tuples with different numbers of fields"
+                                )
                             )
                     else:
-                        raise ValueError("cannot merge tuple with non-tuple record")
+                        raise ak._v2._util.error(
+                            ValueError("cannot merge tuple with non-tuple record")
+                        )
                 elif isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                     pass
                 else:
-                    raise AssertionError(
-                        "cannot merge "
-                        + type(self).__name__
-                        + " with "
-                        + type(array).__name__
+                    raise ak._v2._util.error(
+                        AssertionError(
+                            "cannot merge "
+                            + type(self).__name__
+                            + " with "
+                            + type(array).__name__
+                        )
                     )
 
         else:
@@ -585,6 +624,10 @@ class RecordArray(Content):
             these_fields.sort()
 
             for array in headless:
+                parameters = ak._v2._util.merge_parameters(
+                    parameters, array._parameters, True
+                )
+
                 if isinstance(array, ak._v2.contents.recordarray.RecordArray):
                     if not array.is_tuple:
                         those_fields = array._fields.copy()
@@ -597,27 +640,32 @@ class RecordArray(Content):
                                 trimmed = field[0 : array.length]
                                 for_each_field[i].append(trimmed)
                         else:
-                            raise AssertionError(
-                                "cannot merge records with different sets of field names"
+                            raise ak._v2._util.error(
+                                AssertionError(
+                                    "cannot merge records with different sets of field names"
+                                )
                             )
                     else:
-                        raise AssertionError("cannot merge non-tuple record with tuple")
+                        raise ak._v2._util.error(
+                            AssertionError("cannot merge non-tuple record with tuple")
+                        )
 
                 elif isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                     pass
                 else:
-                    raise AssertionError(
-                        "cannot merge "
-                        + type(self).__name__
-                        + " with "
-                        + type(array).__name__
+                    raise ak._v2._util.error(
+                        AssertionError(
+                            "cannot merge "
+                            + type(self).__name__
+                            + " with "
+                            + type(array).__name__
+                        )
                     )
 
         nextcontents = []
         minlength = None
         for forfield in for_each_field:
-            tail_forfield = forfield[1:]
-            merged = forfield[0].mergemany(tail_forfield)
+            merged = forfield[0].mergemany(forfield[1:])
 
             nextcontents.append(merged)
 
@@ -642,14 +690,16 @@ class RecordArray(Content):
         else:
             return reversed.mergemany(tail[1:])
 
-        raise NotImplementedError(
-            "not implemented: " + type(self).__name__ + " ::mergemany"
+        raise ak._v2._util.error(
+            NotImplementedError(
+                "not implemented: " + type(self).__name__ + " ::mergemany"
+            )
         )
 
-    def fillna(self, value):
+    def fill_none(self, value):
         contents = []
         for content in self._contents:
-            contents.append(content.fillna(value))
+            contents.append(content.fill_none(value))
         return RecordArray(
             contents,
             self._fields,
@@ -659,14 +709,14 @@ class RecordArray(Content):
             self._nplike,
         )
 
-    def _localindex(self, axis, depth):
+    def _local_index(self, axis, depth):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            return self._localindex_axis0()
+            return self._local_index_axis0()
         else:
             contents = []
             for content in self._contents:
-                contents.append(content._localindex(posaxis, depth))
+                contents.append(content._local_index(posaxis, depth))
             return RecordArray(
                 contents,
                 self._fields,
@@ -696,7 +746,7 @@ class RecordArray(Content):
         return True
 
     def _unique(self, negaxis, starts, parents, outlength):
-        raise NotImplementedError
+        raise ak._v2._util.error(NotImplementedError)
 
     def _argsort_next(
         self,
@@ -710,12 +760,7 @@ class RecordArray(Content):
         kind,
         order,
     ):
-        if self._fields is None or len(self._fields) == 0:
-            return ak._v2.contents.NumpyArray(
-                self._nplike.empty(0, np.int64), None, None, self._nplike
-            )
-        else:
-            raise NotImplementedError
+        raise ak._v2._util.error(NotImplementedError)
 
     def _sort_next(
         self, negaxis, starts, parents, outlength, ascending, stable, kind, order
@@ -774,37 +819,30 @@ class RecordArray(Content):
         outlength,
         mask,
         keepdims,
+        behavior,
     ):
-        contents = []
-        for content in self._contents:
-            contents.append(
-                content[: self._length]._reduce_next(
-                    reducer,
-                    negaxis,
-                    starts,
-                    shifts,
-                    parents,
-                    outlength,
-                    mask,
-                    keepdims,
+        reducer_recordclass = ak._v2._util.reducer_recordclass(reducer, self, behavior)
+        if reducer_recordclass is None:
+            raise ak._v2._util.error(
+                TypeError(
+                    "no ak.{} overloads for custom types: {}".format(
+                        reducer.name, ", ".join(self._fields)
+                    )
+                )
+            )
+        else:
+            raise ak._v2._util.error(
+                NotImplementedError(
+                    "overloading reducers for RecordArrays has not been implemented yet"
                 )
             )
 
-        return ak._v2.contents.RecordArray(
-            contents,
-            self._fields,
-            outlength,
-            None,
-            None,
-            self._nplike,
-        )
-
-    def _validityerror(self, path):
+    def _validity_error(self, path):
         for i, cont in enumerate(self.contents):
             if cont.length < self.length:
                 return f'at {path} ("{type(self)}"): len(field({i})) < len(recordarray)'
         for i, cont in enumerate(self.contents):
-            sub = cont.validityerror(f"{path}.field({i})")
+            sub = cont.validity_error(f"{path}.field({i})")
             if sub != "":
                 return sub
         return ""
@@ -817,14 +855,14 @@ class RecordArray(Content):
             result = result + self.identifier._nbytes_part()
         return result
 
-    def _rpad(self, target, axis, depth, clip):
+    def _pad_none(self, target, axis, depth, clip):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            return self.rpad_axis0(target, clip)
+            return self.pad_none_axis0(target, clip)
         else:
             contents = []
             for content in self._contents:
-                contents.append(content._rpad(target, posaxis, depth, clip))
+                contents.append(content._pad_none(target, posaxis, depth, clip))
             if len(contents) == 0:
                 return ak._v2.contents.recordarray.RecordArray(
                     contents,
@@ -863,7 +901,11 @@ class RecordArray(Content):
 
         return pyarrow.Array.from_buffers(
             ak._v2._connect.pyarrow.to_awkwardarrow_type(
-                types, options["extensionarray"], mask_node, self
+                types,
+                options["extensionarray"],
+                options["record_is_scalar"],
+                mask_node,
+                self,
             ),
             length,
             [ak._v2._connect.pyarrow.to_validbits(validbytes)],
@@ -875,7 +917,9 @@ class RecordArray(Content):
             return self._nplike.empty(self.length, dtype=[])
         contents = [x._to_numpy(allow_missing) for x in self._contents]
         if any(len(x.shape) != 1 for x in contents):
-            raise ValueError(f"cannot convert {self} into np.ndarray")
+            raise ak._v2._util.error(
+                ValueError(f"cannot convert {self} into np.ndarray")
+            )
         out = self._nplike.empty(
             contents[0].shape[0],
             dtype=[(str(n), x.dtype) for n, x in zip(self.fields, contents)],
@@ -884,7 +928,7 @@ class RecordArray(Content):
         for n, x in zip(self.fields, contents):
             if isinstance(x, self._nplike.ma.MaskedArray):
                 if mask is None:
-                    mask = self._nplike.ma.zeros(
+                    mask = self._nplike.index_nplike.ma.zeros(
                         self.length, [(n, np.bool_) for n in self.fields]
                     )
                 if x.mask is not None:
@@ -906,28 +950,42 @@ class RecordArray(Content):
             in_function = ""
             if options["function_name"] is not None:
                 in_function = " in " + options["function_name"]
-            raise TypeError(
-                "cannot combine record fields{} unless flatten_records=True".format(
-                    in_function
+            raise ak._v2._util.error(
+                TypeError(
+                    "cannot combine record fields{} unless flatten_records=True".format(
+                        in_function
+                    )
                 )
             )
 
     def _recursively_apply(
-        self, action, depth, depth_context, lateral_context, options
+        self, action, behavior, depth, depth_context, lateral_context, options
     ):
+        if self._nplike.known_shape:
+            contents = [x[: self._length] for x in self._contents]
+        else:
+            contents = self._contents
+
         if options["return_array"]:
 
             def continuation():
+                if not options["allow_records"]:
+                    raise ak._v2._util.error(
+                        ValueError(
+                            f"cannot broadcast records in {options['function_name']}"
+                        )
+                    )
                 return RecordArray(
                     [
                         content._recursively_apply(
                             action,
+                            behavior,
                             depth,
                             copy.copy(depth_context),
                             lateral_context,
                             options,
                         )
-                        for content in self._contents
+                        for content in contents
                     ],
                     self._fields,
                     self._length,
@@ -939,9 +997,10 @@ class RecordArray(Content):
         else:
 
             def continuation():
-                for content in self._contents:
+                for content in contents:
                     content._recursively_apply(
                         action,
+                        behavior,
                         depth,
                         copy.copy(depth_context),
                         lateral_context,
@@ -954,6 +1013,8 @@ class RecordArray(Content):
             depth_context=depth_context,
             lateral_context=lateral_context,
             continuation=continuation,
+            behavior=behavior,
+            nplike=self._nplike,
             options=options,
         )
 
@@ -962,7 +1023,7 @@ class RecordArray(Content):
         elif result is None:
             return continuation()
         else:
-            raise AssertionError(result)
+            raise ak._v2._util.error(AssertionError(result))
 
     def packed(self):
         return RecordArray(
@@ -977,21 +1038,13 @@ class RecordArray(Content):
             self._nplike,
         )
 
-    def _to_list(self, behavior):
-        out = self._to_list_custom(behavior)
+    def _to_list(self, behavior, json_conversions):
+        out = self._to_list_custom(behavior, json_conversions)
         if out is not None:
             return out
 
-        cls = ak._v2._util.recordclass(self, behavior)
-        if cls is not ak._v2.highlevel.Record:
-            length = self._length
-            out = [None] * length
-            for i in range(length):
-                out[i] = cls(self[i])
-            return out
-
-        if self.is_tuple:
-            contents = [x._to_list(behavior) for x in self._contents]
+        if self.is_tuple and json_conversions is None:
+            contents = [x._to_list(behavior, json_conversions) for x in self._contents]
             length = self._length
             out = [None] * length
             for i in range(length):
@@ -1000,7 +1053,9 @@ class RecordArray(Content):
 
         else:
             fields = self._fields
-            contents = [x._to_list(behavior) for x in self._contents]
+            if fields is None:
+                fields = [str(i) for i in range(len(self._contents))]
+            contents = [x._to_list(behavior, json_conversions) for x in self._contents]
             length = self._length
             out = [None] * length
             for i in range(length):
@@ -1018,60 +1073,16 @@ class RecordArray(Content):
             nplike=nplike,
         )
 
-    def _to_json(
-        self,
-        nan_string,
-        infinity_string,
-        minus_infinity_string,
-        complex_real_string,
-        complex_imag_string,
-    ):
-        out = self._to_json_custom()
-        if out is not None:
-            return out
-
-        cls = ak._v2._util.recordclass(self, None)
-        if cls is not ak._v2.highlevel.Record:
-            length = self._length
-            out = [None] * length
-            for i in range(length):
-                out[i] = cls(self[i])
-            return out
-
-        if self.is_tuple:
-            contents = [
-                x._to_json(
-                    nan_string,
-                    infinity_string,
-                    minus_infinity_string,
-                    complex_real_string,
-                    complex_imag_string,
-                )
-                for x in self._contents
-            ]
-            length = self._length
-            out = [None] * length
-            fields = []
-            for i in range(length):
-                fields.append(str(i))
-            for i in range(length):
-                out[i] = dict(zip(fields, [x[i] for x in contents]))
-            return out
-
-        else:
-            fields = self._fields
-            contents = [
-                x._to_json(
-                    nan_string,
-                    infinity_string,
-                    minus_infinity_string,
-                    complex_real_string,
-                    complex_imag_string,
-                )
-                for x in self._contents
-            ]
-            length = self._length
-            out = [None] * length
-            for i in range(length):
-                out[i] = dict(zip(fields, [x[i] for x in contents]))
-            return out
+    def _layout_equal(self, other, index_dtype=True, numpyarray=True):
+        return (
+            self.fields == other.fields
+            and len(self.contents) == len(other.contents)
+            and all(
+                [
+                    self.contents[i].layout_equal(
+                        other.contents[i], index_dtype, numpyarray
+                    )
+                    for i in range(len(self.contents))
+                ]
+            )
+        )
