@@ -15,9 +15,11 @@ class UnmaskedForm(Form):
         form_key=None,
     ):
         if not isinstance(content, Form):
-            raise TypeError(
-                "{} all 'contents' must be Form subclasses, not {}".format(
-                    type(self).__name__, repr(content)
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} all 'contents' must be Form subclasses, not {}".format(
+                        type(self).__name__, repr(content)
+                    )
                 )
             )
 
@@ -53,7 +55,9 @@ class UnmaskedForm(Form):
             return (
                 self._has_identifier == other._has_identifier
                 and self._form_key == other._form_key
-                and _parameters_equal(self._parameters, other._parameters)
+                and _parameters_equal(
+                    self._parameters, other._parameters, only_array_record=True
+                )
                 and self._content == other._content
             )
         else:
@@ -132,6 +136,10 @@ class UnmaskedForm(Form):
     def purelist_depth(self):
         return self._content.purelist_depth
 
+    @property
+    def is_identity_like(self):
+        return self._content.is_identity_like
+
     @property
     def minmax_depth(self):
         return self._content.minmax_depth
@@ -144,6 +152,24 @@ class UnmaskedForm(Form):
     def fields(self):
         return self._content.fields
 
+    @property
+    def is_tuple(self):
+        return self._content.is_tuple
+
     @property
     def dimension_optiontype(self):
         return True
+
+    def _columns(self, path, output, list_indicator):
+        self._content._columns(path, output, list_indicator)
+
+    def _select_columns(self, index, specifier, matches, output):
+        return UnmaskedForm(
+            self._content._select_columns(index, specifier, matches, output),
+            self._has_identifier,
+            self._parameters,
+            self._form_key,
+        )
+
+    def _column_types(self):
+        return self._content._column_types()
